# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['robusta',
 'robusta.api',
 'robusta.cli',
 'robusta.core',
 'robusta.core.discovery',
 'robusta.core.model',
 'robusta.core.persistency',
 'robusta.core.playbooks',
 'robusta.core.playbooks.internal',
 'robusta.core.reporting',
 'robusta.core.schedule',
 'robusta.core.sinks',
 'robusta.core.sinks.datadog',
 'robusta.core.sinks.kafka',
 'robusta.core.sinks.msteams',
 'robusta.core.sinks.opsgenie',
 'robusta.core.sinks.robusta',
 'robusta.core.sinks.robusta.dal',
 'robusta.core.sinks.slack',
 'robusta.core.sinks.telegram',
 'robusta.core.sinks.webhook',
 'robusta.core.triggers',
 'robusta.integrations',
 'robusta.integrations.argocd',
 'robusta.integrations.git',
 'robusta.integrations.kubernetes',
 'robusta.integrations.kubernetes.autogenerated',
 'robusta.integrations.kubernetes.autogenerated.v1',
 'robusta.integrations.kubernetes.autogenerated.v2beta1',
 'robusta.integrations.kubernetes.autogenerated.v2beta2',
 'robusta.integrations.msteams',
 'robusta.integrations.msteams.msteams_elements',
 'robusta.integrations.prometheus',
 'robusta.integrations.resource_analysis',
 'robusta.integrations.scheduled',
 'robusta.integrations.slack',
 'robusta.model',
 'robusta.runner',
 'robusta.utils']

package_data = \
{'': ['*']}

install_requires = \
['click-spinner>=0.1.10,<0.2.0',
 'colorlog>=5.0.1,<6.0.0',
 'cryptography>=36.0.0,<37.0.0',
 'docutils>=0.17.0,<0.18.0',
 'dpath>=2.0.5,<3.0.0',
 'hikaru>=0.5.1-beta.0,<0.6.0',
 'kubernetes>=12.0.1,<13.0.0',
 'markdown2>=2.4.2,<3.0.0',
 'opsgenie-sdk>=2.1.5,<3.0.0',
 'prometheus-client>=0.12.0,<0.13.0',
 'pydantic>=1.8.1,<2.0.0',
 'pymsteams>=0.1.16,<0.2.0',
 'pytz>=2021.3,<2022.0',
 'pyyaml>=6.0,<7.0',
 'slack-sdk>=3.7.0,<4.0.0',
 'toml>=0.10.2,<0.11.0',
 'typer>=0.3.2,<0.4.0',
 'watchgod>=0.7,<0.8',
 'websocket-client>=1.3.2,<2.0.0']

extras_require = \
{':extra == "all"': ['tabulate>=0.8.9,<0.9.0'],
 'all': ['supabase-py>=0.0.2,<0.0.3',
         'Flask>=1.1.2,<2.0.0',
         'grafana-api>=1.0.3,<2.0.0',
         'manhole>=1.8.0,<2.0.0',
         'watchdog>=2.1.0,<3.0.0',
         'better-exceptions>=0.3.3,<0.4.0',
         'CairoSVG>=2.5.2,<3.0.0',
         'kafka-python>=2.0.2,<3.0.0',
         'prometheus-api-client>=0.4.2,<0.5.0',
         'datadog-api-client>=1.2.0,<2.0.0',
         'sentry-sdk>=1.5.2,<2.0.0',
         'dulwich==0.20.28']}

entry_points = \
{'console_scripts': ['robusta = robusta.cli.main:app']}

setup_kwargs = {
    'name': 'robusta-cli',
    'version': '0.9.10',
    'description': '',
    'long_description': None,
    'author': 'Natan Yellin',
    'author_email': 'aantn@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



import copy
import logging
import traceback
from collections import defaultdict
from typing import Any, Dict, Optional, List

from .base_trigger import TriggerEvent, BaseTrigger
from .playbook_utils import merge_global_params
from .playbooks_event_handler import PlaybooksEventHandler
from ..model.events import ExecutionBaseEvent, ExecutionContext
from ..reporting import MarkdownBlock
from ..reporting.base import Finding
from ..reporting.consts import SYNC_RESPONSE_SINK
from ..sinks.robusta.dal.model_conversion import ModelConversion
from ...model.playbook_action import PlaybookAction
from ...model.config import Registry
from .trigger import Trigger
from ...runner.telemetry import Telemetry


class PlaybooksEventHandlerImpl(PlaybooksEventHandler):
    def __init__(self, registry: Registry):
        self.registry = registry

    def handle_trigger(self, trigger_event: TriggerEvent) -> Optional[Dict[str, Any]]:
        playbooks = self.registry.get_playbooks().get_playbooks(trigger_event)
        if not playbooks:  # no registered playbooks for this event type
            return

        execution_response = None
        execution_event: Optional[ExecutionBaseEvent] = None
        sink_findings: Dict[str, List[Finding]] = defaultdict(list)
        for playbook in playbooks:
            fired_trigger = self.__get_fired_trigger(
                trigger_event, playbook.triggers, playbook.get_id()
            )
            if fired_trigger:
                execution_event = fired_trigger.build_execution_event(
                    trigger_event, sink_findings
                )
                if execution_event:  # might not exist for unsupported k8s types
                    execution_event.named_sinks = (
                        playbook.sinks
                        if playbook.sinks is not None
                        else self.registry.get_playbooks().get_default_sinks()
                    )

                    playbook_resp = self.__run_playbook_actions(
                        execution_event,
                        playbook.get_actions(),
                    )
                    if (
                        playbook_resp
                    ):  # For now, only last response applies. (For simplicity reasons)
                        execution_response = playbook_resp
                    if playbook.stop or execution_event.stop_processing:
                        break

        if execution_event:
            self.__handle_findings(execution_event)

        return execution_response

    def run_actions(
        self,
        execution_event: ExecutionBaseEvent,
        actions: List[PlaybookAction],
        sync_response: bool = False,
        no_sinks: bool = False,
    ) -> Optional[Dict[str, Any]]:
        if not no_sinks and execution_event.named_sinks is None:  # take the default sinks only if sinks not excluded
            execution_event.named_sinks = (
                self.registry.get_playbooks().get_default_sinks()
            )

        if sync_response:  # if we need to return sync response, we'll collect the findings under this sink name
            if execution_event.named_sinks:
                execution_event.named_sinks.append(SYNC_RESPONSE_SINK)
            else:
                execution_event.named_sinks = [SYNC_RESPONSE_SINK]

        execution_response = self.__run_playbook_actions(
            execution_event,
            actions,
        )
        self.__handle_findings(execution_event)

        if sync_response:  # add the findings to the response
            execution_response["findings"] = [
                self.__to_finding_json(finding)
                for finding in execution_event.sink_findings[SYNC_RESPONSE_SINK]
            ]

        return execution_response

    def __to_finding_json(self, finding: Finding) -> Dict:
        account_id = self.registry.get_global_config().get("account_id", "")
        cluster_id = self.registry.get_global_config().get("cluster_name", "")
        signing_key = self.registry.get_global_config().get("signing_key", "")

        finding_json = ModelConversion.to_finding_json(account_id, cluster_id, finding)

        finding_json["evidence"] = [
            ModelConversion.to_evidence_json(
                account_id,
                cluster_id,
                SYNC_RESPONSE_SINK,
                signing_key,
                finding.id,
                enrichment
            )
            for enrichment in finding.enrichments
        ]
        return finding_json

    def __prepare_execution_event(self, execution_event: ExecutionBaseEvent):
        execution_event.set_scheduler(self.registry.get_scheduler())
        execution_event.set_context(ExecutionContext(
            account_id=self.registry.get_global_config().get("account_id", ""),
            cluster_name=self.registry.get_global_config().get("cluster_name", "")
        ))

    def run_external_action(
        self,
        action_name: str,
        action_params: Optional[dict],
        sinks: Optional[List[str]],
        sync_response: bool = False,
        no_sinks: bool = False,
    ) -> Optional[Dict[str, Any]]:
        action_def = self.registry.get_actions().get_action(action_name)
        if not action_def:
            return self.__error_resp(f"External action not found {action_name}")

        if not action_def.from_params_func:
            return self.__error_resp(
                f"Action {action_name} cannot run using external event"
            )

        if not no_sinks and sinks:
            if action_params:
                action_params["named_sinks"] = sinks
            else:
                action_params = {"named_sinks": sinks}
        try:
            instantiation_params = action_def.from_params_parameter_class(
                **action_params
            )
        except Exception:
            return self.__error_resp(
                f"Failed to create execution instance for"
                f" {action_name} {action_def.from_params_parameter_class}"
                f" {action_params} {traceback.format_exc()}"
            )

        execution_event = action_def.from_params_func(instantiation_params)
        if not execution_event:
            return self.__error_resp(
                f"Failed to create execution event for "
                f"{action_name} {action_params}"
            )

        playbook_action = PlaybookAction(
            action_name=action_name, action_params=action_params
        )
        return self.run_actions(execution_event, [playbook_action], sync_response, no_sinks)

    @classmethod
    def __error_resp(cls, msg: str) -> dict:
        logging.error(msg)
        return {"success": False, "msg": msg}

    def __run_playbook_actions(
        self,
        execution_event: ExecutionBaseEvent,
        actions: List[PlaybookAction],
    ) -> Dict[str, Any]:
        self.__prepare_execution_event(execution_event)
        execution_event.response = {"success": True}
        for action in actions:
            if execution_event.stop_processing:
                return execution_event.response

            registered_action = self.registry.get_actions().get_action(
                action.action_name
            )

            if (
                not registered_action
            ):  # Might happen if manually trying to trigger incorrect action
                msg = f"action {action.action_name} not found. Skipping for event {type(execution_event)}"
                execution_event.response = self.__error_resp(msg)
                continue

            if not isinstance(execution_event, registered_action.event_type):
                msg = f"Action {action.action_name} requires {registered_action.event_type}"
                execution_event.response = self.__error_resp(msg)
                continue

            if not registered_action.params_type:
                registered_action.func(execution_event)
            else:
                action_params = None
                try:
                    action_params = merge_global_params(
                        self.get_global_config(), action.action_params
                    )
                    params = registered_action.params_type(**action_params)
                except Exception:
                    msg = (
                        f"Failed to create {registered_action.params_type} "
                        f"using {action_params} for running {action.action_name} "
                        f"exc={traceback.format_exc()}"
                    )
                    execution_event.response = self.__error_resp(msg)
                    continue

                try:
                    registered_action.func(execution_event, params)
                except Exception:
                    logging.error(
                        f"Failed to execute action {action.action_name} {action_params}",
                        exc_info=True,
                    )
                    execution_event.add_enrichment(
                        [
                            MarkdownBlock(
                                text=f"Oops... Error processing {action.action_name}"
                            )
                        ]
                    )

        return execution_event.response

    @classmethod
    def __get_fired_trigger(
        cls,
        trigger_event: TriggerEvent,
        playbook_triggers: List[Trigger],
        playbook_id: str,
    ) -> Optional[BaseTrigger]:
        for trigger in playbook_triggers:
            if trigger.get().should_fire(trigger_event, playbook_id):
                return trigger.get()
        return None

    def __handle_findings(self, execution_event: ExecutionBaseEvent):
        sinks_info = self.registry.get_telemetry().sinks_info

        for sink_name in execution_event.named_sinks:
            if SYNC_RESPONSE_SINK == sink_name:
                continue  # not a real sink, just container for findings that needs to be returned synchronously

            for finding in execution_event.sink_findings[sink_name]:
                try:
                    sink = self.registry.get_sinks().sinks.get(sink_name)
                    if not sink:
                        logging.error(
                            f"sink {sink_name} not found. Skipping event finding {finding}"
                        )
                        continue

                    # only write the finding if is matching against the sink matchers
                    if sink.accepts(finding):
                        # create deep copy, so that iterating on one sink enrichments won't affect the others
                        # Each sink has a different findings, but enrichments are shared
                        finding_copy = copy.deepcopy(finding)
                        sink.write_finding(
                            finding_copy, self.registry.get_sinks().platform_enabled
                        )
                        
                        sink_info = sinks_info[sink_name]
                        sink_info.type = sink.__class__.__name__
                        sink_info.findings_count += 1

                except Exception:  # Failure to send to one sink shouldn't fail all
                    logging.error(
                        f"Failed to publish finding to sink {sink_name}", exc_info=True
                    )

    def get_global_config(self) -> dict:
        return self.registry.get_playbooks().get_global_config()

    def get_telemetry(self) -> Telemetry:
        return self.registry.get_telemetry()



import logging
from typing import NamedTuple, Union, Type, List, Optional, Dict
from pydantic.main import BaseModel
from hikaru.model.rel_1_16 import *

from .models import PrometheusKubernetesAlert, PrometheusAlert
from ..helper import exact_match, prefix_match
from ..kubernetes.custom_models import RobustaPod, RobustaDeployment, RobustaJob
from ...core.playbooks.base_trigger import BaseTrigger, TriggerEvent
from ...core.reporting.base import Finding
from ...core.model.events import ExecutionBaseEvent


class PrometheusTriggerEvent(TriggerEvent):
    alert: PrometheusAlert

    def get_event_name(self) -> str:
        return PrometheusTriggerEvent.__name__


class ResourceMapping(NamedTuple):
    hikaru_class: Union[
        Type[RobustaPod], Type[RobustaDeployment], Type[Job], Type[DaemonSet]
    ]
    attribute_name: str
    prometheus_label: str


MAPPINGS = [
    ResourceMapping(RobustaPod, "pod", "pod"),
    ResourceMapping(RobustaDeployment, "deployment", "deployment"),
    ResourceMapping(RobustaJob, "job", "job_name"),
    ResourceMapping(DaemonSet, "daemonset", "daemonset"),
]


class PrometheusAlertTrigger(BaseTrigger):
    alert_name: str = None
    status: str = None
    pod_name_prefix: str = None
    namespace_prefix: str = None
    instance_name_prefix: str = None

    def get_trigger_event(self):
        return PrometheusTriggerEvent.__name__

    def should_fire(self, event: TriggerEvent, playbook_id: str):
        if not isinstance(event, PrometheusTriggerEvent):
            return False

        labels = event.alert.labels
        if not exact_match(self.alert_name, labels["alertname"]):
            return False

        if not exact_match(self.status, event.alert.status):
            return False

        if not prefix_match(self.pod_name_prefix, labels.get("pod")):
            return False

        if not prefix_match(self.namespace_prefix, labels.get("namespace")):
            return False

        if not prefix_match(self.instance_name_prefix, labels.get("instance")):
            return False

        return True

    @classmethod
    def __find_node_by_ip(cls, ip) -> Optional[Node]:
        nodes: NodeList = NodeList.listNode().obj
        for node in nodes.items:
            addresses = [a.address for a in node.status.addresses]
            logging.info(f"node {node.metadata.name} has addresses {addresses}")
            if ip in addresses:
                return node
        return None

    @classmethod
    def __load_node(cls, alert: PrometheusAlert, node_name: str) -> Optional[Node]:
        node = None
        try:
            # sometimes we get an IP:PORT instead of the node name. handle that case
            if ":" in node_name:
                node = cls.__find_node_by_ip(node_name.split(":")[0])
            else:
                node = Node().read(name=node_name)
        except Exception as e:
            logging.info(f"Error loading Node kubernetes object {alert}. error: {e}")

        return node

    def build_execution_event(
        self, event: PrometheusTriggerEvent, sink_findings: Dict[str, List[Finding]]
    ) -> Optional[ExecutionBaseEvent]:
        labels = event.alert.labels
        execution_event = PrometheusKubernetesAlert(
            sink_findings=sink_findings,
            alert=event.alert,
            alert_name=labels["alertname"],
            alert_severity=labels.get("severity"),
        )

        namespace = labels.get("namespace", "default")

        for mapping in MAPPINGS:
            try:
                resource_name = labels.get(mapping.prometheus_label, None)
                if not resource_name:
                    continue
                resource = mapping.hikaru_class().read(
                    name=resource_name, namespace=namespace
                )
                setattr(execution_event, mapping.attribute_name, resource)
                logging.info(
                    f"Successfully loaded Kubernetes resource {resource_name} for alert {execution_event.alert_name}"
                )
            except Exception as e:
                logging.info(
                    f"Error loading {mapping.hikaru_class} kubernetes object {execution_event.alert}. error: {e}"
                )

        node_name = labels.get("node")
        if node_name:
            execution_event.node = self.__load_node(execution_event.alert, node_name)

        # we handle nodes differently than other resources
        node_name = labels.get("instance", None)
        job_name = labels.get(
            "job", None
        )  # a prometheus "job" not a kubernetes "job" resource
        # when the job_name is kube-state-metrics "instance" refers to the IP of kube-state-metrics not the node
        # If the alert has pod, the 'instance' attribute contains the pod ip
        if not execution_event.node and node_name and job_name != "kube-state-metrics":
            execution_event.node = self.__load_node(execution_event.alert, node_name)

        return execution_event

    @staticmethod
    def get_execution_event_type() -> type:
        return PrometheusKubernetesAlert


class PrometheusAlertTriggers(BaseModel):
    on_prometheus_alert: Optional[PrometheusAlertTrigger]



from typing import List

from .playbook_scheduler import PlaybooksScheduler
from ...model.playbook_definition import PlaybookDefinition


class PlaybooksSchedulerManager(PlaybooksScheduler):
    def update(self, playbooks: List[PlaybookDefinition]):
        """Update the scheduler with the new deployed playbooks"""
        pass



# This file was autogenerated. Do not edit.

import logging
import traceback
from dataclasses import dataclass
from abc import abstractmethod
from hikaru.model import Pod,ReplicaSet,DaemonSet,Deployment,StatefulSet,Service,Event,HorizontalPodAutoscaler,Node,ClusterRole,ClusterRoleBinding,Job,Namespace,ServiceAccount,PersistentVolume
from hikaru.utils import Response
from pydantic import BaseModel
from typing import Union, Optional, List
from ..base_event import K8sBaseChangeEvent
from ....core.model.events import ExecutionBaseEvent, ExecutionEventBaseParams
from ..custom_models import RobustaPod,RobustaDeployment,RobustaJob
from hikaru.model.rel_1_16.v1 import ClusterRole as v1ClusterRole    
from hikaru.model.rel_1_16.v1 import ClusterRoleBinding as v1ClusterRoleBinding    
from hikaru.model.rel_1_16.v1 import DaemonSet as v1DaemonSet    
from hikaru.model.rel_1_16.v1 import Deployment as v1Deployment    
from hikaru.model.rel_1_16.v1 import Event as v1Event    
from hikaru.model.rel_1_16.v1 import HorizontalPodAutoscaler as v1HorizontalPodAutoscaler    
from hikaru.model.rel_1_16.v1 import Job as v1Job    
from hikaru.model.rel_1_16.v1 import Namespace as v1Namespace    
from hikaru.model.rel_1_16.v1 import Node as v1Node    
from hikaru.model.rel_1_16.v1 import PersistentVolume as v1PersistentVolume    
from hikaru.model.rel_1_16.v1 import Pod as v1Pod    
from hikaru.model.rel_1_16.v1 import ReplicaSet as v1ReplicaSet    
from hikaru.model.rel_1_16.v1 import Service as v1Service    
from hikaru.model.rel_1_16.v1 import ServiceAccount as v1ServiceAccount    
from hikaru.model.rel_1_16.v1 import StatefulSet as v1StatefulSet    
from hikaru.model.rel_1_16.v2beta1 import ClusterRole as v2beta1ClusterRole    
from hikaru.model.rel_1_16.v2beta1 import ClusterRoleBinding as v2beta1ClusterRoleBinding    
from hikaru.model.rel_1_16.v2beta1 import DaemonSet as v2beta1DaemonSet    
from hikaru.model.rel_1_16.v2beta1 import Deployment as v2beta1Deployment    
from hikaru.model.rel_1_16.v2beta1 import Event as v2beta1Event    
from hikaru.model.rel_1_16.v2beta1 import HorizontalPodAutoscaler as v2beta1HorizontalPodAutoscaler    
from hikaru.model.rel_1_16.v2beta1 import Job as v2beta1Job    
from hikaru.model.rel_1_16.v2beta1 import Namespace as v2beta1Namespace    
from hikaru.model.rel_1_16.v2beta1 import Node as v2beta1Node    
from hikaru.model.rel_1_16.v2beta1 import PersistentVolume as v2beta1PersistentVolume    
from hikaru.model.rel_1_16.v2beta1 import Pod as v2beta1Pod    
from hikaru.model.rel_1_16.v2beta1 import ReplicaSet as v2beta1ReplicaSet    
from hikaru.model.rel_1_16.v2beta1 import Service as v2beta1Service    
from hikaru.model.rel_1_16.v2beta1 import ServiceAccount as v2beta1ServiceAccount    
from hikaru.model.rel_1_16.v2beta1 import StatefulSet as v2beta1StatefulSet    
from hikaru.model.rel_1_16.v2beta2 import ClusterRole as v2beta2ClusterRole    
from hikaru.model.rel_1_16.v2beta2 import ClusterRoleBinding as v2beta2ClusterRoleBinding    
from hikaru.model.rel_1_16.v2beta2 import DaemonSet as v2beta2DaemonSet    
from hikaru.model.rel_1_16.v2beta2 import Deployment as v2beta2Deployment    
from hikaru.model.rel_1_16.v2beta2 import Event as v2beta2Event    
from hikaru.model.rel_1_16.v2beta2 import HorizontalPodAutoscaler as v2beta2HorizontalPodAutoscaler    
from hikaru.model.rel_1_16.v2beta2 import Job as v2beta2Job    
from hikaru.model.rel_1_16.v2beta2 import Namespace as v2beta2Namespace    
from hikaru.model.rel_1_16.v2beta2 import Node as v2beta2Node    
from hikaru.model.rel_1_16.v2beta2 import PersistentVolume as v2beta2PersistentVolume    
from hikaru.model.rel_1_16.v2beta2 import Pod as v2beta2Pod    
from hikaru.model.rel_1_16.v2beta2 import ReplicaSet as v2beta2ReplicaSet    
from hikaru.model.rel_1_16.v2beta2 import Service as v2beta2Service    
from hikaru.model.rel_1_16.v2beta2 import ServiceAccount as v2beta2ServiceAccount    
from hikaru.model.rel_1_16.v2beta2 import StatefulSet as v2beta2StatefulSet    


LOADERS_MAPPINGS = {
    'pod': (True, RobustaPod.readNamespacedPod),
    'replicaset': (True, ReplicaSet.readNamespacedReplicaSet),
    'daemonset': (True, DaemonSet.readNamespacedDaemonSet),
    'deployment': (True, RobustaDeployment.readNamespacedDeployment),
    'statefulset': (True, StatefulSet.readNamespacedStatefulSet),
    'service': (True, Service.readNamespacedService),
    'event': (True, Event.readNamespacedEvent),
    'horizontalpodautoscaler': (True, HorizontalPodAutoscaler.readNamespacedHorizontalPodAutoscaler),
    'node': (False, Node.readNode),
    'clusterrole': (False, ClusterRole.readClusterRole),
    'clusterrolebinding': (False, ClusterRoleBinding.readClusterRoleBinding),
    'job': (True, RobustaJob.readNamespacedJob),
    'namespace': (False, Namespace.readNamespace),
    'serviceaccount': (True, ServiceAccount.readNamespacedServiceAccount),
    'persistentvolume': (False, PersistentVolume.readPersistentVolume),
}


class ResourceLoader:
    @staticmethod
    def read_resource(kind: str, name: str, namespace: str = None) -> Response:
        resource_mapper = LOADERS_MAPPINGS[kind.lower()]
        if not resource_mapper:
            raise Exception("resource loader not found")

        if resource_mapper[0]:  # namespaced resource
            return resource_mapper[1](name=name, namespace=namespace)
        else:
            return resource_mapper[1](name=name)


class ResourceAttributes(ExecutionEventBaseParams):
    kind: str
    name: str
    namespace: Optional[str] = None


@dataclass
class KubernetesResourceEvent(ExecutionBaseEvent):
    _obj: Optional[Union[RobustaPod,ReplicaSet,DaemonSet,RobustaDeployment,StatefulSet,Service,Event,HorizontalPodAutoscaler,Node,ClusterRole,ClusterRoleBinding,RobustaJob,Namespace,ServiceAccount,PersistentVolume]] = None

    def __init__(self, obj: Union[RobustaPod,ReplicaSet,DaemonSet,RobustaDeployment,StatefulSet,Service,Event,HorizontalPodAutoscaler,Node,ClusterRole,ClusterRoleBinding,RobustaJob,Namespace,ServiceAccount,PersistentVolume], named_sinks: List[str]):
        super().__init__(named_sinks=named_sinks)
        self._obj = obj

    def get_resource(self) -> Optional[Union[RobustaPod,ReplicaSet,DaemonSet,RobustaDeployment,StatefulSet,Service,Event,HorizontalPodAutoscaler,Node,ClusterRole,ClusterRoleBinding,RobustaJob,Namespace,ServiceAccount,PersistentVolume]]:
        return self._obj

    @staticmethod
    def from_params(params: ResourceAttributes) -> Optional["KubernetesResourceEvent"]:
        try:
            obj = ResourceLoader.read_resource(
                kind=params.kind, 
                name=params.name, 
                namespace=params.namespace
            ).obj
        except Exception:
            logging.error(f"Could not load resource {params}", exc_info=True)
            return None
        return KubernetesResourceEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class KubernetesAnyChangeEvent(K8sBaseChangeEvent):
    obj: Optional[Union[RobustaDeployment,RobustaJob,RobustaPod,v1ClusterRole,v1ClusterRoleBinding,v1DaemonSet,v1Event,v1HorizontalPodAutoscaler,v1Namespace,v1Node,v1PersistentVolume,v1ReplicaSet,v1Service,v1ServiceAccount,v1StatefulSet,v2beta1ClusterRole,v2beta1ClusterRoleBinding,v2beta1DaemonSet,v2beta1Event,v2beta1HorizontalPodAutoscaler,v2beta1Namespace,v2beta1Node,v2beta1PersistentVolume,v2beta1ReplicaSet,v2beta1Service,v2beta1ServiceAccount,v2beta1StatefulSet,v2beta2ClusterRole,v2beta2ClusterRoleBinding,v2beta2DaemonSet,v2beta2Event,v2beta2HorizontalPodAutoscaler,v2beta2Namespace,v2beta2Node,v2beta2PersistentVolume,v2beta2ReplicaSet,v2beta2Service,v2beta2ServiceAccount,v2beta2StatefulSet]] = None
    old_obj: Optional[Union[RobustaDeployment,RobustaJob,RobustaPod,v1ClusterRole,v1ClusterRoleBinding,v1DaemonSet,v1Event,v1HorizontalPodAutoscaler,v1Namespace,v1Node,v1PersistentVolume,v1ReplicaSet,v1Service,v1ServiceAccount,v1StatefulSet,v2beta1ClusterRole,v2beta1ClusterRoleBinding,v2beta1DaemonSet,v2beta1Event,v2beta1HorizontalPodAutoscaler,v2beta1Namespace,v2beta1Node,v2beta1PersistentVolume,v2beta1ReplicaSet,v2beta1Service,v2beta1ServiceAccount,v2beta1StatefulSet,v2beta2ClusterRole,v2beta2ClusterRoleBinding,v2beta2DaemonSet,v2beta2Event,v2beta2HorizontalPodAutoscaler,v2beta2Namespace,v2beta2Node,v2beta2PersistentVolume,v2beta2ReplicaSet,v2beta2Service,v2beta2ServiceAccount,v2beta2StatefulSet]] = None

    def get_resource(self) -> Optional[Union[RobustaDeployment,RobustaJob,RobustaPod,v1ClusterRole,v1ClusterRoleBinding,v1DaemonSet,v1Event,v1HorizontalPodAutoscaler,v1Namespace,v1Node,v1PersistentVolume,v1ReplicaSet,v1Service,v1ServiceAccount,v1StatefulSet,v2beta1ClusterRole,v2beta1ClusterRoleBinding,v2beta1DaemonSet,v2beta1Event,v2beta1HorizontalPodAutoscaler,v2beta1Namespace,v2beta1Node,v2beta1PersistentVolume,v2beta1ReplicaSet,v2beta1Service,v2beta1ServiceAccount,v2beta1StatefulSet,v2beta2ClusterRole,v2beta2ClusterRoleBinding,v2beta2DaemonSet,v2beta2Event,v2beta2HorizontalPodAutoscaler,v2beta2Namespace,v2beta2Node,v2beta2PersistentVolume,v2beta2ReplicaSet,v2beta2Service,v2beta2ServiceAccount,v2beta2StatefulSet]]:
        return self.obj


class PodAttributes(ExecutionEventBaseParams):
    name: str
    namespace: str


@dataclass
class PodEvent(KubernetesResourceEvent):
    def __init__(self, obj: RobustaPod, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_pod(self) -> Optional[RobustaPod]:
        return self._obj

    @staticmethod
    def from_params(params: PodAttributes) -> Optional["PodEvent"]:
        try:
            obj = RobustaPod.readNamespacedPod(name=params.name, namespace=params.namespace).obj
        except Exception:
            logging.error(f"Could not load Pod {params}", exc_info=True)
            return None
        return PodEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class PodChangeEvent(PodEvent, KubernetesAnyChangeEvent):
    obj: Optional[RobustaPod] = None
    old_obj: Optional[RobustaPod] = None

    def get_pod(self) -> Optional[RobustaPod]:
        return self.obj


class ReplicaSetAttributes(ExecutionEventBaseParams):
    name: str
    namespace: str


@dataclass
class ReplicaSetEvent(KubernetesResourceEvent):
    def __init__(self, obj: ReplicaSet, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_replicaset(self) -> Optional[ReplicaSet]:
        return self._obj

    @staticmethod
    def from_params(params: ReplicaSetAttributes) -> Optional["ReplicaSetEvent"]:
        try:
            obj = ReplicaSet.readNamespacedReplicaSet(name=params.name, namespace=params.namespace).obj
        except Exception:
            logging.error(f"Could not load ReplicaSet {params}", exc_info=True)
            return None
        return ReplicaSetEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class ReplicaSetChangeEvent(ReplicaSetEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1ReplicaSet,v2beta1ReplicaSet,v2beta2ReplicaSet]] = None
    old_obj: Optional[Union[v1ReplicaSet,v2beta1ReplicaSet,v2beta2ReplicaSet]] = None

    def get_replicaset(self) -> Optional[Union[v1ReplicaSet,v2beta1ReplicaSet,v2beta2ReplicaSet]]:
        return self.obj


class DaemonSetAttributes(ExecutionEventBaseParams):
    name: str
    namespace: str


@dataclass
class DaemonSetEvent(KubernetesResourceEvent):
    def __init__(self, obj: DaemonSet, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_daemonset(self) -> Optional[DaemonSet]:
        return self._obj

    @staticmethod
    def from_params(params: DaemonSetAttributes) -> Optional["DaemonSetEvent"]:
        try:
            obj = DaemonSet.readNamespacedDaemonSet(name=params.name, namespace=params.namespace).obj
        except Exception:
            logging.error(f"Could not load DaemonSet {params}", exc_info=True)
            return None
        return DaemonSetEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class DaemonSetChangeEvent(DaemonSetEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1DaemonSet,v2beta1DaemonSet,v2beta2DaemonSet]] = None
    old_obj: Optional[Union[v1DaemonSet,v2beta1DaemonSet,v2beta2DaemonSet]] = None

    def get_daemonset(self) -> Optional[Union[v1DaemonSet,v2beta1DaemonSet,v2beta2DaemonSet]]:
        return self.obj


class DeploymentAttributes(ExecutionEventBaseParams):
    name: str
    namespace: str


@dataclass
class DeploymentEvent(KubernetesResourceEvent):
    def __init__(self, obj: RobustaDeployment, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_deployment(self) -> Optional[RobustaDeployment]:
        return self._obj

    @staticmethod
    def from_params(params: DeploymentAttributes) -> Optional["DeploymentEvent"]:
        try:
            obj = RobustaDeployment.readNamespacedDeployment(name=params.name, namespace=params.namespace).obj
        except Exception:
            logging.error(f"Could not load Deployment {params}", exc_info=True)
            return None
        return DeploymentEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class DeploymentChangeEvent(DeploymentEvent, KubernetesAnyChangeEvent):
    obj: Optional[RobustaDeployment] = None
    old_obj: Optional[RobustaDeployment] = None

    def get_deployment(self) -> Optional[RobustaDeployment]:
        return self.obj


class StatefulSetAttributes(ExecutionEventBaseParams):
    name: str
    namespace: str


@dataclass
class StatefulSetEvent(KubernetesResourceEvent):
    def __init__(self, obj: StatefulSet, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_statefulset(self) -> Optional[StatefulSet]:
        return self._obj

    @staticmethod
    def from_params(params: StatefulSetAttributes) -> Optional["StatefulSetEvent"]:
        try:
            obj = StatefulSet.readNamespacedStatefulSet(name=params.name, namespace=params.namespace).obj
        except Exception:
            logging.error(f"Could not load StatefulSet {params}", exc_info=True)
            return None
        return StatefulSetEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class StatefulSetChangeEvent(StatefulSetEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1StatefulSet,v2beta1StatefulSet,v2beta2StatefulSet]] = None
    old_obj: Optional[Union[v1StatefulSet,v2beta1StatefulSet,v2beta2StatefulSet]] = None

    def get_statefulset(self) -> Optional[Union[v1StatefulSet,v2beta1StatefulSet,v2beta2StatefulSet]]:
        return self.obj


class ServiceAttributes(ExecutionEventBaseParams):
    name: str
    namespace: str


@dataclass
class ServiceEvent(KubernetesResourceEvent):
    def __init__(self, obj: Service, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_service(self) -> Optional[Service]:
        return self._obj

    @staticmethod
    def from_params(params: ServiceAttributes) -> Optional["ServiceEvent"]:
        try:
            obj = Service.readNamespacedService(name=params.name, namespace=params.namespace).obj
        except Exception:
            logging.error(f"Could not load Service {params}", exc_info=True)
            return None
        return ServiceEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class ServiceChangeEvent(ServiceEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1Service,v2beta1Service,v2beta2Service]] = None
    old_obj: Optional[Union[v1Service,v2beta1Service,v2beta2Service]] = None

    def get_service(self) -> Optional[Union[v1Service,v2beta1Service,v2beta2Service]]:
        return self.obj


class EventAttributes(ExecutionEventBaseParams):
    name: str
    namespace: str


@dataclass
class EventEvent(KubernetesResourceEvent):
    def __init__(self, obj: Event, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_event(self) -> Optional[Event]:
        return self._obj

    @staticmethod
    def from_params(params: EventAttributes) -> Optional["EventEvent"]:
        try:
            obj = Event.readNamespacedEvent(name=params.name, namespace=params.namespace).obj
        except Exception:
            logging.error(f"Could not load Event {params}", exc_info=True)
            return None
        return EventEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class EventChangeEvent(EventEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1Event,v2beta1Event,v2beta2Event]] = None
    old_obj: Optional[Union[v1Event,v2beta1Event,v2beta2Event]] = None

    def get_event(self) -> Optional[Union[v1Event,v2beta1Event,v2beta2Event]]:
        return self.obj


class HorizontalPodAutoscalerAttributes(ExecutionEventBaseParams):
    name: str
    namespace: str


@dataclass
class HorizontalPodAutoscalerEvent(KubernetesResourceEvent):
    def __init__(self, obj: HorizontalPodAutoscaler, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_horizontalpodautoscaler(self) -> Optional[HorizontalPodAutoscaler]:
        return self._obj

    @staticmethod
    def from_params(params: HorizontalPodAutoscalerAttributes) -> Optional["HorizontalPodAutoscalerEvent"]:
        try:
            obj = HorizontalPodAutoscaler.readNamespacedHorizontalPodAutoscaler(name=params.name, namespace=params.namespace).obj
        except Exception:
            logging.error(f"Could not load HorizontalPodAutoscaler {params}", exc_info=True)
            return None
        return HorizontalPodAutoscalerEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class HorizontalPodAutoscalerChangeEvent(HorizontalPodAutoscalerEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1HorizontalPodAutoscaler,v2beta1HorizontalPodAutoscaler,v2beta2HorizontalPodAutoscaler]] = None
    old_obj: Optional[Union[v1HorizontalPodAutoscaler,v2beta1HorizontalPodAutoscaler,v2beta2HorizontalPodAutoscaler]] = None

    def get_horizontalpodautoscaler(self) -> Optional[Union[v1HorizontalPodAutoscaler,v2beta1HorizontalPodAutoscaler,v2beta2HorizontalPodAutoscaler]]:
        return self.obj


class NodeAttributes(ExecutionEventBaseParams):
    name: str



@dataclass
class NodeEvent(KubernetesResourceEvent):
    def __init__(self, obj: Node, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_node(self) -> Optional[Node]:
        return self._obj

    @staticmethod
    def from_params(params: NodeAttributes) -> Optional["NodeEvent"]:
        try:
            obj = Node.readNode(name=params.name).obj
        except Exception:
            logging.error(f"Could not load Node {params}", exc_info=True)
            return None
        return NodeEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class NodeChangeEvent(NodeEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1Node,v2beta1Node,v2beta2Node]] = None
    old_obj: Optional[Union[v1Node,v2beta1Node,v2beta2Node]] = None

    def get_node(self) -> Optional[Union[v1Node,v2beta1Node,v2beta2Node]]:
        return self.obj


class ClusterRoleAttributes(ExecutionEventBaseParams):
    name: str



@dataclass
class ClusterRoleEvent(KubernetesResourceEvent):
    def __init__(self, obj: ClusterRole, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_clusterrole(self) -> Optional[ClusterRole]:
        return self._obj

    @staticmethod
    def from_params(params: ClusterRoleAttributes) -> Optional["ClusterRoleEvent"]:
        try:
            obj = ClusterRole.readClusterRole(name=params.name).obj
        except Exception:
            logging.error(f"Could not load ClusterRole {params}", exc_info=True)
            return None
        return ClusterRoleEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class ClusterRoleChangeEvent(ClusterRoleEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1ClusterRole,v2beta1ClusterRole,v2beta2ClusterRole]] = None
    old_obj: Optional[Union[v1ClusterRole,v2beta1ClusterRole,v2beta2ClusterRole]] = None

    def get_clusterrole(self) -> Optional[Union[v1ClusterRole,v2beta1ClusterRole,v2beta2ClusterRole]]:
        return self.obj


class ClusterRoleBindingAttributes(ExecutionEventBaseParams):
    name: str



@dataclass
class ClusterRoleBindingEvent(KubernetesResourceEvent):
    def __init__(self, obj: ClusterRoleBinding, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_clusterrolebinding(self) -> Optional[ClusterRoleBinding]:
        return self._obj

    @staticmethod
    def from_params(params: ClusterRoleBindingAttributes) -> Optional["ClusterRoleBindingEvent"]:
        try:
            obj = ClusterRoleBinding.readClusterRoleBinding(name=params.name).obj
        except Exception:
            logging.error(f"Could not load ClusterRoleBinding {params}", exc_info=True)
            return None
        return ClusterRoleBindingEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class ClusterRoleBindingChangeEvent(ClusterRoleBindingEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1ClusterRoleBinding,v2beta1ClusterRoleBinding,v2beta2ClusterRoleBinding]] = None
    old_obj: Optional[Union[v1ClusterRoleBinding,v2beta1ClusterRoleBinding,v2beta2ClusterRoleBinding]] = None

    def get_clusterrolebinding(self) -> Optional[Union[v1ClusterRoleBinding,v2beta1ClusterRoleBinding,v2beta2ClusterRoleBinding]]:
        return self.obj


class JobAttributes(ExecutionEventBaseParams):
    name: str
    namespace: str


@dataclass
class JobEvent(KubernetesResourceEvent):
    def __init__(self, obj: RobustaJob, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_job(self) -> Optional[RobustaJob]:
        return self._obj

    @staticmethod
    def from_params(params: JobAttributes) -> Optional["JobEvent"]:
        try:
            obj = RobustaJob.readNamespacedJob(name=params.name, namespace=params.namespace).obj
        except Exception:
            logging.error(f"Could not load Job {params}", exc_info=True)
            return None
        return JobEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class JobChangeEvent(JobEvent, KubernetesAnyChangeEvent):
    obj: Optional[RobustaJob] = None
    old_obj: Optional[RobustaJob] = None

    def get_job(self) -> Optional[RobustaJob]:
        return self.obj


class NamespaceAttributes(ExecutionEventBaseParams):
    name: str



@dataclass
class NamespaceEvent(KubernetesResourceEvent):
    def __init__(self, obj: Namespace, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_namespace(self) -> Optional[Namespace]:
        return self._obj

    @staticmethod
    def from_params(params: NamespaceAttributes) -> Optional["NamespaceEvent"]:
        try:
            obj = Namespace.readNamespace(name=params.name).obj
        except Exception:
            logging.error(f"Could not load Namespace {params}", exc_info=True)
            return None
        return NamespaceEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class NamespaceChangeEvent(NamespaceEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1Namespace,v2beta1Namespace,v2beta2Namespace]] = None
    old_obj: Optional[Union[v1Namespace,v2beta1Namespace,v2beta2Namespace]] = None

    def get_namespace(self) -> Optional[Union[v1Namespace,v2beta1Namespace,v2beta2Namespace]]:
        return self.obj


class ServiceAccountAttributes(ExecutionEventBaseParams):
    name: str
    namespace: str


@dataclass
class ServiceAccountEvent(KubernetesResourceEvent):
    def __init__(self, obj: ServiceAccount, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_serviceaccount(self) -> Optional[ServiceAccount]:
        return self._obj

    @staticmethod
    def from_params(params: ServiceAccountAttributes) -> Optional["ServiceAccountEvent"]:
        try:
            obj = ServiceAccount.readNamespacedServiceAccount(name=params.name, namespace=params.namespace).obj
        except Exception:
            logging.error(f"Could not load ServiceAccount {params}", exc_info=True)
            return None
        return ServiceAccountEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class ServiceAccountChangeEvent(ServiceAccountEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1ServiceAccount,v2beta1ServiceAccount,v2beta2ServiceAccount]] = None
    old_obj: Optional[Union[v1ServiceAccount,v2beta1ServiceAccount,v2beta2ServiceAccount]] = None

    def get_serviceaccount(self) -> Optional[Union[v1ServiceAccount,v2beta1ServiceAccount,v2beta2ServiceAccount]]:
        return self.obj


class PersistentVolumeAttributes(ExecutionEventBaseParams):
    name: str



@dataclass
class PersistentVolumeEvent(KubernetesResourceEvent):
    def __init__(self, obj: PersistentVolume, named_sinks: List[str]):
        super().__init__(obj=obj, named_sinks=named_sinks)

    def get_persistentvolume(self) -> Optional[PersistentVolume]:
        return self._obj

    @staticmethod
    def from_params(params: PersistentVolumeAttributes) -> Optional["PersistentVolumeEvent"]:
        try:
            obj = PersistentVolume.readPersistentVolume(name=params.name).obj
        except Exception:
            logging.error(f"Could not load PersistentVolume {params}", exc_info=True)
            return None
        return PersistentVolumeEvent(obj=obj, named_sinks=params.named_sinks)


@dataclass
class PersistentVolumeChangeEvent(PersistentVolumeEvent, KubernetesAnyChangeEvent):
    obj: Optional[Union[v1PersistentVolume,v2beta1PersistentVolume,v2beta2PersistentVolume]] = None
    old_obj: Optional[Union[v1PersistentVolume,v2beta1PersistentVolume,v2beta2PersistentVolume]] = None

    def get_persistentvolume(self) -> Optional[Union[v1PersistentVolume,v2beta1PersistentVolume,v2beta2PersistentVolume]]:
        return self.obj



KIND_TO_EVENT_CLASS = {
    'pod': PodChangeEvent,
    'replicaset': ReplicaSetChangeEvent,
    'daemonset': DaemonSetChangeEvent,
    'deployment': DeploymentChangeEvent,
    'statefulset': StatefulSetChangeEvent,
    'service': ServiceChangeEvent,
    'event': EventChangeEvent,
    'horizontalpodautoscaler': HorizontalPodAutoscalerChangeEvent,
    'node': NodeChangeEvent,
    'clusterrole': ClusterRoleChangeEvent,
    'clusterrolebinding': ClusterRoleBindingChangeEvent,
    'job': JobChangeEvent,
    'namespace': NamespaceChangeEvent,
    'serviceaccount': ServiceAccountChangeEvent,
    'persistentvolume': PersistentVolumeChangeEvent
}



from enum import Enum

SYNC_RESPONSE_SINK = "robusta-synchronized-response-sink"


class FindingType(Enum):
    ISSUE = "issue"
    CONF_CHANGE = "configuration_change"
    HEALTH_CHECK = "health_check"
    REPORT = "report"


# Finding sources
class FindingSource(Enum):
    NONE = None  # empty default
    KUBERNETES_API_SERVER = "kubernetes_api_server"
    PROMETHEUS = "prometheus"
    MANUAL = "manual"
    CALLBACK = "callback"


# Finding subject types
class FindingSubjectType(Enum):
    TYPE_NONE = None  # empty default
    TYPE_DEPLOYMENT = "deployment"
    TYPE_NODE = "node"
    TYPE_POD = "pod"
    TYPE_JOB = "job"
    TYPE_DAEMONSET = "daemonset"

    @staticmethod
    def from_kind(kind: str):
        if kind == "deployment":
            return FindingSubjectType.TYPE_DEPLOYMENT
        elif kind == "node":
            return FindingSubjectType.TYPE_NODE
        elif kind == "pod":
            return FindingSubjectType.TYPE_POD
        elif kind == "job":
            return FindingSubjectType.TYPE_JOB
        elif kind == "daemonset":
            return FindingSubjectType.TYPE_DAEMONSET
        return FindingSubjectType.TYPE_NONE


# Annotations
class SlackAnnotations:
    UNFURL = "unfurl"
    ATTACHMENT = "attachment"



import json
import os
import sys

import typer
from pydantic.main import BaseModel

ROBUSTA_BACKEND_PROFILE = os.environ.get("ROBUSTA_BACKEND_PROFILE", "")


class BackendProfile(BaseModel):
    robusta_cloud_api_host: str = ""
    robusta_ui_domain: str = ""
    robusta_relay_ws_address: str = ""
    robusta_relay_external_actions_url: str = ""
    robusta_telemetry_endpoint: str = ""
    robusta_store_token_url: str = ""
    custom_profile: bool = False


# default values
backend_profile = BackendProfile(
    robusta_cloud_api_host="https://api.robusta.dev",
    robusta_ui_domain="https://platform.robusta.dev",
    robusta_store_token_url="https://api.robusta.dev/auth/server/tokens"
)

if ROBUSTA_BACKEND_PROFILE:
    typer.secho(
        f"Using Robusta backend profile: {ROBUSTA_BACKEND_PROFILE}",
        color="blue",
    )
    backend_profile = BackendProfile.parse_file(ROBUSTA_BACKEND_PROFILE)
    backend_profile.custom_profile = True

    profile_dict = backend_profile.dict()
    for attribute, val in profile_dict.items():
        if not profile_dict.get(attribute):
            typer.secho(f"Illegal profile. Missing {attribute}. Aborting!")
            sys.exit(1)



import inspect
import hashlib

# this returns a hash that will change if the function's definition changes
# this is desirable because it lets you warn users if a playbook is called with parameters that
# were generated by a previous version of the playbook. in that case you can ask the user if you should
# call the playbook anyway even though the code that the callback refers to has changed
import json

from pydantic import BaseModel


def get_function_hash(func):
    plaintext = str(inspect.getfullargspec(func)).encode() + func.__code__.co_code
    return hashlib.sha256(plaintext).hexdigest()


def action_hash(func, action_params: BaseModel, additional_data: dict) -> str:
    hash_input = (
        f"{get_function_hash(func)}"
        + ("None" if additional_data is None else json.dumps(additional_data))
        + ("None" if action_params is None else action_params.json())
    )
    return hashlib.md5(hash_input.encode()).hexdigest()



import base64
import os
import time
import uuid

import requests
import typer
from pydantic import BaseModel
from collections import namedtuple

from .backend_profile import backend_profile
from .utils import log_title

app = typer.Typer()

SLACK_INTEGRATION_SERVICE_ADDRESS = os.environ.get(
    "SLACK_INTEGRATION_SERVICE_ADDRESS",
    f"{backend_profile.robusta_cloud_api_host}/integrations/slack/get-token",
)
SlackApiKey = namedtuple("SlackApiKey", "key team_name")


def wait_for_slack_api_key(id: str) -> SlackApiKey:
    while True:
        try:
            response_json = requests.get(
                f"{SLACK_INTEGRATION_SERVICE_ADDRESS}?id={id}"
            ).json()
            if response_json["token"]:
                return SlackApiKey(
                    str(response_json["token"]), response_json.get("team-name", None)
                )
            time.sleep(0.5)
        except Exception as e:
            log_title(f"Error getting slack token {e}")


def _get_slack_key_once() -> SlackApiKey:
    id = str(uuid.uuid4())
    url = f"{backend_profile.robusta_cloud_api_host}/integrations/slack?id={id}"
    typer.secho(
        f"If your browser does not automatically launch, open the below url:\n{url}"
    )
    typer.launch(url)
    slack_api_key = wait_for_slack_api_key(id)
    return slack_api_key


def get_slack_key() -> (str, str):
    slack_api_key = _get_slack_key_once()
    if not slack_api_key or not slack_api_key.team_name:
        return slack_api_key.key, ""
    team_name = slack_api_key.team_name
    team_name_styled = typer.style(team_name, fg=typer.colors.CYAN, bold=True)
    typer.secho(f"You've just connected Robusta to the Slack of: {team_name_styled}")
    return slack_api_key.key, team_name_styled


@app.command()
def slack():
    """generate slack api key"""
    key, workspace = get_slack_key()
    log_title(
        f"Connected to Slack workspace {workspace}.\n"
        f"Your Slack key is:\n{key}\nAdd it to the slack sink configuration"
    )


if __name__ == "__main__":
    app()



from tabulate import tabulate

from .telegram_client import TelegramClient
from .telegram_sink_params import TelegramSinkConfigWrapper
from ..transformer import Transformer
from ...reporting.blocks import (
    MarkdownBlock,
    TableBlock,
    FileBlock,
)
from ...reporting.base import (
    Finding,
    FindingSeverity, BaseBlock,
)
from ..sink_base import SinkBase

SEVERITY_EMOJI_MAP = {
    FindingSeverity.INFO: u"\U0001F7E2",
    FindingSeverity.LOW: u"\U0001F7E1",
    FindingSeverity.MEDIUM: u"\U0001F7E0",
    FindingSeverity.HIGH: u"\U0001F534",
}
INVESTIGATE_ICON = u"\U0001F50E"


class TelegramSink(SinkBase):
    def __init__(self, sink_config: TelegramSinkConfigWrapper, registry):
        super().__init__(sink_config.telegram_sink, registry)

        self.client = TelegramClient(sink_config.telegram_sink.chat_id, sink_config.telegram_sink.bot_token)
        self.send_files = sink_config.telegram_sink.send_files

    def write_finding(self, finding: Finding, platform_enabled: bool):
        self.__send_telegram_message(finding, platform_enabled)

    def __send_telegram_message(self, finding: Finding, platform_enabled: bool):
        self.client.send_message(self.__get_message_text(finding, platform_enabled))
        if self.send_files:
            for enrichment in finding.enrichments:
                file_blocks = [block for block in enrichment.blocks if isinstance(block, FileBlock)]
                for block in file_blocks:
                    self.client.send_file(file_name=block.filename, contents=block.contents)
                table_blocks = [block for block in enrichment.blocks if isinstance(block, TableBlock)]
                for block in table_blocks:
                    table_text = tabulate(block.render_rows(), headers=block.headers, tablefmt="presto")
                    table_name = block.table_name if block.table_name else "table"
                    self.client.send_file(
                        file_name=f"{table_name}.txt",
                        contents=table_text.encode("utf-8")
                    )

    def __get_message_text(self, finding: Finding, platform_enabled: bool):
        message_content = self.__build_telegram_title(finding.title, finding.severity)

        if platform_enabled:
            message_content += f"[{INVESTIGATE_ICON} Investigate]({finding.investigate_uri})\n\n"

        blocks = [MarkdownBlock(text=f"*Source:* `{self.cluster_name}`\n\n")]

        # first add finding description block
        if finding.description:
            blocks.append(MarkdownBlock(finding.description))

        for enrichment in finding.enrichments:
            blocks.extend([block for block in enrichment.blocks if self.__is_telegram_text_block(block)])

        for block in blocks:
            block_text = Transformer.to_standard_markdown([block])
            if len(block_text) + len(message_content) >= 4096:  # telegram message size limit
                break
            message_content += block_text + "\n"

        return message_content

    @classmethod
    def __is_telegram_text_block(cls, block: BaseBlock) -> bool:
        # enrichments text tables are too big for mobile device
        return not ( isinstance(block, FileBlock) or isinstance(block, TableBlock) )

    @classmethod
    def __build_telegram_title(cls, title: str, severity: FindingSeverity) -> str:
        icon = SEVERITY_EMOJI_MAP.get(severity, "")
        return f"{icon} **{severity.name} - {title}**\n\n"




import logging
import os
import re

from typing import Dict, Optional
from pydantic.types import SecretStr


def get_env_replacement(value: str) -> Optional[str]:
    env_values = re.findall(r"{{[ ]*env\.(.*)[ ]*}}", value)
    if env_values:
        env_var_value = os.environ.get(env_values[0].strip(), None)
        if not env_var_value:
            msg = f"ENV var replacement {env_values[0]} does not exist for param: {value}"
            logging.error(msg)
            raise Exception(msg)
        return env_var_value
    return None


def replace_env_vars_values(values: Dict) -> Dict:
    for key, value in values.items():
        if isinstance(value, str):
            env_var_value = get_env_replacement(value)
            if env_var_value:
                values[key] = env_var_value
        elif isinstance(value, SecretStr):
            env_var_value = get_env_replacement(value.get_secret_value())
            if env_var_value:
                values[key] = SecretStr(env_var_value)

    return values


def merge_global_params(global_config: dict, config_params: dict) -> dict:
    merged = global_config.copy()
    merged.update(config_params)
    return merged



import logging

from .msteams_msg import MsTeamsMsg
from ...core.reporting import (
    Enrichment,
    BaseBlock,
    MarkdownBlock,
    DividerBlock,
    HeaderBlock,
    TableBlock,
    ListBlock,
    KubernetesDiffBlock,
    CallbackBlock,
    FileBlock, Finding,
)


class MsTeamsSender:

    @classmethod
    def __to_ms_teams(cls, block: BaseBlock, msg: MsTeamsMsg):
        if isinstance(block, MarkdownBlock):
            msg.markdown_block(block)
        elif isinstance(block, DividerBlock):
            msg.divider_block()
        elif isinstance(block, HeaderBlock):
            msg.header_block(block)
        elif isinstance(block, TableBlock):
            msg.table(block)
        elif isinstance(block, ListBlock):
            msg.items_list(block)
        elif isinstance(block, KubernetesDiffBlock):
            msg.diff(block)
        elif isinstance(block, CallbackBlock):
            logging.error(
                f"CallbackBlock not supported for msteams"
            )
        else:
            logging.error(
                f"cannot convert block of type {type(block)} to msteams format block: {block}"
            )

    @classmethod
    def __split_block_to_files_and_all_the_rest(cls, enrichment : Enrichment):
        files_blocks = []
        other_blocks = []

        for block in enrichment.blocks:
            if isinstance(block, FileBlock):
                files_blocks.append(block)
            else:
                other_blocks.append(block)
        return files_blocks, other_blocks

    @classmethod
    def send_finding_to_ms_teams(cls, webhook_url: str, finding: Finding, platform_enabled: bool, cluster_name: str):
        msg = MsTeamsMsg(webhook_url)
        msg.write_title_and_desc(finding.title, finding.description, finding.severity.name,
                                 platform_enabled, finding.investigate_uri, cluster_name)
        
        for enrichment in finding.enrichments:
            files_blocks, other_blocks = cls.__split_block_to_files_and_all_the_rest(enrichment)
            for block in other_blocks:
                cls.__to_ms_teams(block, msg)

            msg.upload_files(files_blocks)
            msg.write_current_section()

        msg.send()



import time
from pydantic import BaseModel

from robusta.core.reporting.action_requests import (
    sign_action_request,
    ExternalActionRequest,
    ActionRequestBody,
)
from . import CallbackChoice
from ..playbooks.actions_registry import Action


class ExternalActionRequestBuilder(BaseModel):
    @classmethod
    def create_for_func(
        cls,
        choice: CallbackChoice,
        sink: str,
        text: str,
        account_id: str,
        cluster_name: str,
        signing_key: str,
    ):
        if choice.action is None:
            raise Exception(
                f"The callback for choice {text} is None. Did you accidentally pass `foo()` as a callback and not `foo`?"
            )
        if not Action.is_action(choice.action):
            raise Exception(
                f"{choice.action} is not a function that was decorated with @action"
            )

        if not signing_key:
            raise Exception(
                f"Cannot create callback request with no signing key. Configure signing_key in globalConfig"
            )
        action_params = (
            {} if choice.action_params is None else choice.action_params.dict()
        )
        if choice.kubernetes_object:
            action_params["kind"] = choice.kubernetes_object.kind
            action_params["name"] = choice.kubernetes_object.metadata.name
            if hasattr(choice.kubernetes_object.metadata, "namespace"):
                action_params["namespace"] = choice.kubernetes_object.metadata.namespace

        body = ActionRequestBody(
            account_id=account_id,
            cluster_name=cluster_name,
            timestamp=time.time(),
            action_name=choice.action.__name__,
            action_params=action_params,
            sinks=[sink],
            origin="callback",
        )
        return ExternalActionRequest(
            body=body,
            signature=sign_action_request(body, signing_key),
        )



import json
import logging
from kafka import KafkaProducer

from .kafka_sink_params import KafkaSinkConfigWrapper
from ...reporting.blocks import KubernetesDiffBlock, JsonBlock
from ...reporting.base import Finding, Enrichment
from ..sink_base import SinkBase


class KafkaSink(SinkBase):
    def __init__(self, sink_config: KafkaSinkConfigWrapper, registry):
        super().__init__(sink_config.kafka_sink, registry)
        self.producer = KafkaProducer(
            bootstrap_servers=sink_config.kafka_sink.kafka_url
        )
        self.topic = sink_config.kafka_sink.topic

    def write_finding(self, finding: Finding, platform_enabled: bool):
        for enrichment in finding.enrichments:
            self.send_enrichment(
                enrichment,
                finding.subject.name,
                finding.subject.subject_type.value,
                finding.subject.namespace,
            )

    def send_enrichment(
        self,
        enrichment: Enrichment,
        resource_name: str,
        resource_type: str,
        resource_namespace: str,
    ):
        kafka_blocks = [
            block
            for block in enrichment.blocks
            if (isinstance(block, KubernetesDiffBlock) or isinstance(block, JsonBlock))
        ]
        if not kafka_blocks:  # currently supporting sending kafka sink only diffs
            if len(enrichment.blocks) > 0:
                logging.warning(
                    "Kafka sink unsupported enrichment types. Currently only diff/json enrichment is supported"
                )
            return

        message_payload = ""
        for block in kafka_blocks:
            if isinstance(block, KubernetesDiffBlock):
                data = {
                    "cluster_name": self.cluster_name,
                    "resource_name": resource_name,
                    "resource_namespace": resource_namespace,
                    "resource_type": resource_type,
                    "message": f"{resource_type} properties change",
                    "changed_properties": [
                        {
                            "property": ".".join(attribute_diff.path),
                            "old": attribute_diff.other_value,
                            "new": attribute_diff.value,
                        }
                        for attribute_diff in block.diffs
                    ],
                }
                message_payload = json.dumps(data).encode("utf-8")
            elif isinstance(block, JsonBlock):
                json_obj = json.loads(block.json_str)
                json_obj["cluster_name"] = self.cluster_name
                message_payload = json.dumps(json_obj).encode("utf-8")

            self.producer.send(self.topic, value=message_payload)



import logging
import os
from typing import Optional
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey

from ..core.model.env_vars import RSA_KEYS_PATH


class AuthProvider:

    def __init__(self):
        logging.info(f"Loading RSA keys from {RSA_KEYS_PATH}")
        self.prv: RSAPrivateKey = self.__class__._load_private_key(os.path.join(RSA_KEYS_PATH, "prv"))
        self.pub: RSAPublicKey = self.__class__._load_public_key(os.path.join(RSA_KEYS_PATH, "pub"))

    def get_private_rsa_key(self) -> RSAPrivateKey:
        return self.prv

    def get_public_rsa_key(self) -> RSAPublicKey:
        return self.pub

    @staticmethod
    def _load_private_key(file_name: str) -> Optional[RSAPrivateKey]:
        try:
            with open(file_name, "rb") as key_file:
                private_key = serialization.load_pem_private_key(
                    key_file.read(),
                    password=None,
                )
                logging.info(f"Loaded private key file {file_name}")
                return private_key
        except Exception:
            logging.error(f"Could not load private key file {file_name}")

        return None

    @staticmethod
    def _load_public_key(file_name: str) -> Optional[RSAPublicKey]:
        try:
            with open(file_name, "rb") as key_file:
                public_key = serialization.load_pem_public_key(key_file.read())
                logging.info(f"Loaded public key file {file_name}")
                return public_key
        except Exception:
            logging.error(f"Could not load public key file {file_name}")

        return None



import logging
from datetime import datetime
from flask import Flask, request, jsonify
from werkzeug.middleware.dispatcher import DispatcherMiddleware
from prometheus_client import make_wsgi_app

from .config_loader import ConfigLoader
from ..integrations.prometheus.trigger import PrometheusTriggerEvent
from ..integrations.kubernetes.base_triggers import (
    IncomingK8sEventPayload,
    K8sTriggerEvent,
)
from ..core.playbooks.playbooks_event_handler import PlaybooksEventHandler
from ..integrations.prometheus.models import AlertManagerEvent
from ..core.model.env_vars import NUM_EVENT_THREADS
from ..utils.task_queue import TaskQueue, QueueMetrics

app = Flask(__name__)
app.wsgi_app = DispatcherMiddleware(app.wsgi_app, {"/metrics": make_wsgi_app()})


class Web:
    api_server_queue: TaskQueue
    alerts_queue: TaskQueue
    event_handler: PlaybooksEventHandler
    metrics: QueueMetrics
    loader: ConfigLoader


    @staticmethod
    def init(event_handler: PlaybooksEventHandler, loader: ConfigLoader):
        Web.metrics = QueueMetrics()
        Web.api_server_queue = TaskQueue(
            name="api server queue", num_workers=NUM_EVENT_THREADS, metrics=Web.metrics
        )
        Web.alerts_queue = TaskQueue(
            name="alerts queue", num_workers=NUM_EVENT_THREADS, metrics=Web.metrics
        )
        Web.event_handler = event_handler
        Web.loader = loader


    @staticmethod
    def run():
        app.run(host="0.0.0.0", use_reloader=False)

    @staticmethod
    @app.route("/api/alerts", methods=["POST"])
    def handle_alert_event():
        alert_manager_event = AlertManagerEvent(**request.get_json())
        for alert in alert_manager_event.alerts:
            Web.alerts_queue.add_task(
                Web.event_handler.handle_trigger, PrometheusTriggerEvent(alert=alert)
            )
        
        Web.event_handler.get_telemetry().last_alert_at = str(datetime.now())
        return jsonify(success=True)

    @staticmethod
    @app.route("/api/handle", methods=["POST"])
    def handle_api_server_event():
        k8s_payload = IncomingK8sEventPayload(**request.get_json()["data"])
        Web.api_server_queue.add_task(
            Web.event_handler.handle_trigger, K8sTriggerEvent(k8s_payload=k8s_payload)
        )
        return jsonify(success=True)

    @staticmethod
    @app.route("/api/trigger", methods=["POST"])
    def handle_manual_trigger():
        data = request.get_json()
        if not data.get("action_name", None):
            msg = f"Illegal trigger request {data}"
            logging.error(msg)
            return {"success": False, "msg": msg}

        return jsonify(
            Web.event_handler.run_external_action(
                action_name=data["action_name"],
                action_params=data.get("action_params", None),
                sinks=data.get("sinks", None),
            )
        )

    @staticmethod
    @app.route("/api/playbooks/reload", methods=["POST"])
    def handle_playbooks_reload():
        Web.loader.reload("reload request")
        return jsonify(success=True)



from datetime import datetime, timedelta, tzinfo
from typing import Optional

from prometheus_api_client import PrometheusConnect

from ..prometheus.utils import PrometheusDiscovery
from ...core.model.env_vars import PROMETHEUS_REQUEST_TIMEOUT_SECONDS


class MemoryAnalyzer:
    def __init__(self, prometheus_url: str, prometheus_tzinfo: tzinfo):
        if prometheus_url is None or prometheus_url == "":
            prometheus_url = PrometheusDiscovery.find_prometheus_url()

        self.prom = PrometheusConnect(url=prometheus_url, disable_ssl=True)
        self.default_prometheus_params = {"timeout": PROMETHEUS_REQUEST_TIMEOUT_SECONDS}

        self.prometheus_tzinfo = prometheus_tzinfo

    def get_max_node_memory_usage_in_percentage(self, node_name: str, duration: timedelta) -> Optional[float]:
        """
        Gets the maximal memory usage (in percentage) for the node with the given node name, in the time range between now and now - duration.
        :return: a float between 0 and 1, representing the maximal percentage of memory in use by the given node
        """

        max_memory_usage_in_percentage = self._get_max_value_in_first_series_of_query(
            f"instance:node_memory_utilisation:ratio{{job=\"node-exporter\", job=\"node-exporter\", instance=\"{node_name}\"}}",
            duration
        )
        return max_memory_usage_in_percentage

    def get_container_max_memory_usage_in_bytes(self, node_name: str, pod_name: str, container_name: str, duration: timedelta) -> float:
        """
        Returns the maximal memory usage (in bytes) for the given container, in the time range between now and now - duration.r
        """

        # We use the container_memory_usage_bytes metric, with the following filters:
        # - node={node_name}:
        #   In order to filter on the given node.
        #
        # - pod={pod_name}:
        #   In order to filter on the given pod.
        #
        # - image!="":
        #   In order to ignore the metric of the pod in general, which sums the memory of each of its containers.
        #
        # - container!="POD":
        #   In order to ignore paused containers. For more information, please see:
        #   https://stackoverflow.com/questions/68683403/what-is-the-container-pod-label-in-prometheus-and-why-do-most-examples-exclude
        #
        # - container={container_name}:
        #   In order to filter on the given container name.
        #
        # - id=~"/kubepods/.*":
        #   In order to filter multiple ids that are returned for the same container.
        #   In my minikube, for example, 2 time series are returned for each container, one with id=/docker/* and one with id=/kubepods/*.
        #   However, an id of the form /kubepods/* always exists. Also, the following prefixes are always mutual exclusive:
        #       /kubepods/burstable/*
        #       /kubepods/besteffort/*
        #       /kubepods/guaranteed/*
        #   Therefore, there will always be exactly one id of the form /kubepods/* prefix for each container.
        #   See https://stackoverflow.com/questions/49035724/how-do-i-resolve-kubepods-besteffort-poduuid-to-a-pod-name for more information.

        return self._get_max_value_in_first_series_of_query(
            f"container_memory_usage_bytes{{node=\"{node_name}\", pod=\"{pod_name}\", image!=\"\", "
            f"container!=\"POD\", container=\"{container_name}\", id=~\"/kubepods/.*\"}}",
            duration
        )

    def _get_max_value_in_first_series_of_query(self, promql_query: str, duration: timedelta) -> Optional[float]:
        results = self._query(promql_query, duration)

        if len(results) == 0:
            return None

        series = results[0]
        series_values = series["values"]
        max_value_in_series = max([float(val) for (timestamp, val) in series_values])
        return max_value_in_series

    def _query(self, promql_query: str, duration: timedelta) -> list:
        end_time = datetime.now(tz=self.prometheus_tzinfo)
        start_time = end_time - duration
        step = 1
        results = self.prom.custom_query_range(
            promql_query,
            start_time,
            end_time,
            step,
            {
                "timeout": self.default_prometheus_params["timeout"]
            }
        )

        return results


k8s_memory_factors = {
    "K": 1000,
    "M": 1000*1000,
    "G": 1000*1000*1000,
    "P": 1000*1000*1000*1000,
    "E": 1000*1000*1000*1000*1000,
    "Ki": 1024,
    "Mi": 1024*1024,
    "Gi": 1024*1024*1024,
    "Pi": 1024*1024*1024*1024,
    "Ei": 1024*1024*1024*1024*1024
}


class K8sMemoryTransformer:
    @staticmethod
    def get_number_of_bytes_from_kubernetes_mem_spec(mem_spec: str) -> int:
        if len(mem_spec) > 2 and mem_spec[-2:] in k8s_memory_factors:
            return int(mem_spec[:-2]) * k8s_memory_factors[mem_spec[-2:]]

        if len(mem_spec) > 1 and mem_spec[-1] in k8s_memory_factors:
            return int(mem_spec[:-1]) * k8s_memory_factors[mem_spec[-1]]

        raise Exception("number of bytes could not be extracted from memory spec: " + mem_spec)



from pydantic.main import BaseModel
from typing import Dict


class NodeInfo(BaseModel):
    name: str
    node_creation_time: str
    internal_ip: str
    external_ip: str
    taints: str
    conditions: str
    memory_capacity: int  # MB
    memory_allocatable: int  # MB
    memory_allocated: int  # MB
    cpu_capacity: float
    cpu_allocatable: float
    cpu_allocated: float
    pods_count: int
    pods: str
    node_info: Dict
    deleted: bool = False

    def __compare_node_info(self, other_node_info):
        return self.node_info == other_node_info

    def __eq__(self, other):
        if not isinstance(other, NodeInfo):
            return NotImplemented

        ignored_fields = ["deleted"]
        filtered_self = {k: v for k, v in self.dict().items() if k not in ignored_fields}
        filtered_other = {k: v for k, v in other.dict().items() if k not in ignored_fields}
        return filtered_self == filtered_other



import inspect
import json
import logging
from pydantic import BaseModel
from pydantic.fields import ModelField
from .docs import Docstring, DocstringField


class DocumentedModel(BaseModel):
    """
    Extends pydantic.BaseModel so that you can document models with docstrings and not using
        Field(..., description="foo")

    You write docs in the docstring and behind the scenes the actual Fields()  will be updated
    This way pydantic's introspection and schema-generation works like normal and includes those docs

    """

    # warning: __init_subclass__ only works on Python 3.6 and above
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        docs = inspect.getdoc(cls)
        if docs is not None:
            cls.__update_fields_from_docstring(docs)

    @classmethod
    def __update_fields_from_docstring(cls, docstring):
        """
        Updates pydantic fields according to the docstring so that:

        1. you can document individual fields with ":var fieldname: description" in the model's docstring
        2. you can provide examples for individual fields with ":example fieldname: value" in the model's docstring
        3. docs about individual fields (like :var: and :example:) are removed from the root docs
        """
        docs = Docstring(docstring)
        for doc_field in docs.fields:
            if doc_field.field_target not in cls.__fields__:
                logging.warning(
                    f"Docstring contains description of {doc_field.field_target} which is not a field on the model"
                )
                continue

            f: ModelField = cls.__fields__[doc_field.field_target]
            if doc_field.field_type == "example":
                f.field_info.extra["example"] = cls.__parse_example(
                    doc_field.field_value
                )
            if doc_field.field_type == "var":
                if f.field_info.description:
                    logging.warning(
                        f"Overriding existing field description '{f.field_info.description}' with '{doc_field.field_value}'"
                    )
                f.field_info.description = doc_field.field_value
        cls.__doc__ = docs.description

    @staticmethod
    def __parse_example(example: str):
        try:
            return json.loads(example)
        except json.JSONDecodeError:
            return example



# TODO: add a KubernetesBlock for rendering Kubernetes object in a standard way
# Notes on how we define all the classes below:
# 1. We use pydantic and not dataclasses so that field types are validated
# 2. We add __init__ methods ourselves for convenience. Without our own __init__ method, something like
#       HeaderBlock("foo") doesn't work. Only HeaderBlock(text="foo") would be allowed by pydantic.
import textwrap
from copy import deepcopy
from typing import List, Callable, Dict, Any, Iterable, Sequence, Optional

import hikaru
from hikaru import DiffDetail, DiffType
from hikaru.model import HikaruDocumentBase
from pydantic import BaseModel
from tabulate import tabulate

from .custom_rendering import render_value
from .base import BaseBlock
from ..model.env_vars import PRINTED_TABLE_MAX_WIDTH

BLOCK_SIZE_LIMIT = 2997  # due to slack block size limit of 3000


class MarkdownBlock(BaseBlock):
    """
    A Block of `Markdown <https://en.wikipedia.org/wiki/Markdown>`__
    """

    text: str

    def __init__(self, text: str, dedent: bool = False):
        """
        :param text: one or more paragraphs of Markdown markup
        :param dedent: if True, remove common indentation so that you can use multi-line docstrings.
        """
        if dedent:
            if text[0] == "\n":
                text = text[1:]
            text = textwrap.dedent(text)

        if len(text) >= BLOCK_SIZE_LIMIT:
            text = text[:BLOCK_SIZE_LIMIT] + "..."
        super().__init__(text=text)


class DividerBlock(BaseBlock):
    """
    A visual separator between other blocks
    """

    pass


class FileBlock(BaseBlock):
    """
    A file of any type. Used for images, log files, binary files, and more.
    """

    filename: str
    contents: bytes

    def __init__(self, filename: str, contents: bytes):
        """
        :param filename: the file's name
        :param contents: the file's contents
        """
        super().__init__(filename=filename, contents=contents)


class HeaderBlock(BaseBlock):
    """
    Text formatted as a header
    """

    text: str

    def __init__(self, text: str):
        """
        :param text: the header
        """
        super().__init__(text=text)


class ListBlock(BaseBlock):
    """
    A list of items, nicely formatted
    """

    items: List[str]

    def __init__(self, items: List[str]):
        """
        :param items: a list of strings
        """
        super().__init__(items=items)

    def to_markdown(self) -> MarkdownBlock:
        mrkdwn = [f"* {item}" for item in self.items]
        return MarkdownBlock("\n".join(mrkdwn))


# TODO: we should add a generalization of this which isn't K8s specific
class KubernetesDiffBlock(BaseBlock):
    """
    A diff between two versions of a Kubernetes object
    """

    diffs: List[DiffDetail]
    old: Optional[str]
    new: Optional[str]
    resource_name: Optional[str]
    num_additions: Optional[int]
    num_deletions: Optional[int]
    num_modifications: Optional[int]

    # note that interesting_diffs might be a subset of the full diff between old and new
    def __init__(
        self,
        interesting_diffs: List[DiffDetail],
        old: Optional[HikaruDocumentBase],
        new: Optional[HikaruDocumentBase],
        name: str,
        namespace: str = None
    ):
        """
        :param interesting_diffs: parts of the diff to emphasize - some sinks will only show these to save space
        :param old: the old version of the object
        :param new: the new version of the object
        """
        num_additions = len(
            [d for d in interesting_diffs if d.diff_type == DiffType.ADDED]
        )
        num_deletions = len(
            [d for d in interesting_diffs if d.diff_type == DiffType.REMOVED]
        )
        num_modifications = len(interesting_diffs) - num_additions - num_deletions

        resource_name = self._obj_to_name(old, name, namespace) or self._obj_to_name(new, name, namespace)

        super().__init__(
            diffs=interesting_diffs,
            old=self._obj_to_content(old),
            new=self._obj_to_content(new),
            resource_name=resource_name,
            num_additions=num_additions,
            num_deletions=num_deletions,
            num_modifications=num_modifications,
        )

    @staticmethod
    def _obj_to_content(obj: Optional[HikaruDocumentBase]):
        if obj is None:
            return ""
        else:
            return hikaru.get_yaml(obj)

    @staticmethod
    def _obj_to_name(obj: Optional[HikaruDocumentBase], name: str, namespace: str = ""):
        if obj is None:
            return ""

        kind = getattr(obj, "kind", "").lower()
        obj_name = ""
        if kind:
            obj_name += f"{kind}/"
        if namespace:
            obj_name += f"{namespace}/"

        return f"{obj_name}{name}.yaml"


class JsonBlock(BaseBlock):
    """
    Json data
    """

    json_str: str

    def __init__(self, json_str: str):
        """
        :param json_str: json as a string
        """
        super().__init__(json_str=json_str)


class TableBlock(BaseBlock):
    """
    Table display of a list of lists.

    Note: Wider tables appears as a file attachment on Slack, because they aren't rendered properly inline
    """

    rows: List[List]
    headers: Sequence[str] = ()
    column_renderers: Dict = {}
    table_name: str = ""

    def __init__(
        self, rows: List[List], headers: Sequence[str] = (), column_renderers: Dict = {}, table_name: str = ""
    ):
        """
        :param rows: a list of rows. each row is a list of columns
        :param headers: names of each column
        """
        super().__init__(rows=rows, headers=headers, column_renderers=column_renderers, table_name=table_name)

    @classmethod
    def __calc_max_width(cls, headers, rendered_rows, table_max_width: int) -> List[int]:
        # We need to make sure the total table width, doesn't exceed the max width,
        # otherwise, the table is printed corrupted
        columns_max_widths = [len(header) for header in headers]
        for row in rendered_rows:
            for idx, val in enumerate(row):
                columns_max_widths[idx] = max(len(str(val)), columns_max_widths[idx])

        if (
            sum(columns_max_widths) > table_max_width
        ):  # We want to limit the widest column
            largest_width = max(columns_max_widths)
            widest_column_idx = columns_max_widths.index(largest_width)
            diff = sum(columns_max_widths) - table_max_width
            columns_max_widths[widest_column_idx] = largest_width - diff
            if (
                columns_max_widths[widest_column_idx] < 0
            ):  # in case the diff is bigger than the largest column
                # just divide equally
                columns_max_widths = [
                    int(table_max_width / len(columns_max_widths))
                    for i in range(0, len(columns_max_widths))
                ]

        return columns_max_widths

    @classmethod
    def __to_strings_rows(cls, rows):
        # This is just to assert all row column values are strings. Tabulate might fail on other types
        return [list(map(lambda column_value: str(column_value), row)) for row in rows]

    def to_markdown(self) -> MarkdownBlock:
        table_header = f"{self.table_name}\n" if self.table_name else ""
        return MarkdownBlock(f"{table_header}```\n{self.to_table_string()}\n```")

    def to_table_string(self, table_max_width: int = PRINTED_TABLE_MAX_WIDTH) -> str:
        rendered_rows = self.__to_strings_rows(self.render_rows())
        col_max_width = self.__calc_max_width(self.headers, rendered_rows, table_max_width)
        return tabulate(
            rendered_rows,
            headers=self.headers,
            tablefmt="presto",
            maxcolwidths=col_max_width,
        )

    def render_rows(self) -> List[List]:
        if self.column_renderers is None:
            return self.rows
        new_rows = deepcopy(self.rows)
        for (column_name, renderer_type) in self.column_renderers.items():
            column_idx = self.headers.index(column_name)
            for row in new_rows:
                row[column_idx] = render_value(renderer_type, row[column_idx])
        return new_rows


class KubernetesFieldsBlock(TableBlock):
    """
    A nicely formatted Kubernetes objects, with a subset of the fields shown
    """

    def __init__(
        self,
        k8s_obj: HikaruDocumentBase,
        fields: List[str],
        explanations: Dict[str, str] = {},
    ):
        """
        :param k8s_obj: a kubernetes object
        :param fields: a list of fields to display. for example ["metadata.name", "metadata.namespace"]
        :param explanations: an explanation for each field. for example {"metadata.name": "the pods name"}
        """
        if explanations:
            rows = [
                [f, k8s_obj.object_at_path(f.split(".")), explanations.get(f, "")]
                for f in fields
            ]
            super().__init__(rows=rows, headers=["field", "value", "explanation"])
        else:
            rows = [[f, k8s_obj.object_at_path(f.split("."))] for f in fields]
            super().__init__(rows=rows, headers=["field", "value"])


class CallbackChoice(BaseModel):
    action: Callable
    action_params: Optional[BaseModel]
    kubernetes_object: Optional[Any]

    class Config:
        arbitrary_types_allowed = True


class CallbackBlock(BaseBlock):
    """
    A set of buttons that allows callbacks from the sink - for example, a button in Slack that will trigger another action when clicked
    """

    choices: Dict[str, CallbackChoice]

    def __init__(self, choices: Dict[str, CallbackChoice]):
        """
        :param choices: a dict mapping between each the text on each button to the action it triggers
        """
        super().__init__(choices=choices)



from .msteams_base import MsTeamsBase
from ..msteams_mark_down_fix_url import MsTeamsMarkDownFixUrl


class MsTeamsTextBlock(MsTeamsBase):
    def __init__(self,
                 text : str,
                 is_subtle : bool = None,
                 wrap: bool = None,
                 weight: str = None,
                 is_visible : bool = True,
                 separator : bool = False,
                 font_size : str = 'medium',
                 horizontal_alignment : str = "left"
                 ):
        super().__init__(self.__text_block(text, is_subtle, wrap, weight, is_visible,
                                           separator, font_size, horizontal_alignment))

    def __to_msteams_text(self, text: str) -> str:
        teams_text = MsTeamsMarkDownFixUrl().fix_text(text)
        teams_text = teams_text.replace("```", "")
        return teams_text

    def __text_block(self,
                     text: str,
                     is_subtle: bool = None,
                     wrap: bool = None,
                     weight: str = None,
                     is_visible: bool = True,
                     separator: bool = False,
                     font_size: str = 'medium',
                     horizontal_alignment: str = "left"
                     ):
        self.block = {
            "type": "TextBlock",
            "text": self.__to_msteams_text(text),
            "size": font_size,
            "wrap": True,
        }

        if not is_visible:
            self.block["isVisible"] = is_visible

        if separator:
            self.block["separator"] = separator

        if horizontal_alignment != "left":
            self.block["horizontalAlignment"] = horizontal_alignment

        if is_subtle is not None:
            self.block["isSubtle"] = is_subtle

        if weight is not None:
            self.block["weight"] = weight

        return self.block

    def get_text_from_block(self) -> str:
        return self.block["text"]

    def set_text_from_block(self, text: str):
        self.block["text"] = text





import re
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Dict, Any
from urllib.parse import urlparse, unquote_plus
from pydantic import BaseModel

from hikaru.model import Node, DaemonSet

from ..kubernetes.autogenerated.events import (
    PodEvent,
    NodeEvent,
    DaemonSetEvent,
    JobEvent,
    DeploymentEvent,
)
from ...core.reporting import (
    Finding,
    FindingSubjectType,
    FindingSubject,
    FindingSeverity,
    FindingSource,
)
from ..kubernetes.custom_models import RobustaPod, RobustaDeployment, RobustaJob

SEVERITY_MAP = {
    "critical": FindingSeverity.HIGH,
    "error": FindingSeverity.MEDIUM,
    "warning": FindingSeverity.LOW,
    "info": FindingSeverity.INFO,
}


# for parsing incoming data
class PrometheusAlert(BaseModel):
    endsAt: datetime
    generatorURL: str
    startsAt: datetime
    fingerprint: Optional[str] = ""
    status: str
    labels: Dict[Any, Any]
    annotations: Dict[Any, Any]


# for parsing incoming data
class AlertManagerEvent(BaseModel):
    alerts: List[PrometheusAlert] = []
    externalURL: str
    groupKey: str
    version: str
    commonAnnotations: Optional[Dict[Any, Any]] = None
    commonLabels: Optional[Dict[Any, Any]] = None
    groupLabels: Optional[Dict[Any, Any]] = None
    receiver: str
    status: str


# everything here needs to be optional due to annoying subtleties regarding dataclass inheritance
# see explanation in the code for BaseEvent
@dataclass
class PrometheusKubernetesAlert(
    PodEvent, NodeEvent, DeploymentEvent, JobEvent, DaemonSetEvent
):
    alert: Optional[PrometheusAlert] = None
    alert_name: Optional[str] = None
    alert_severity: Optional[str] = None
    node: Optional[Node] = None
    pod: Optional[RobustaPod] = None
    deployment: Optional[RobustaDeployment] = None
    job: Optional[RobustaJob] = None
    daemonset: Optional[DaemonSet] = None

    def get_node(self) -> Optional[Node]:
        return self.node

    def get_pod(self) -> Optional[RobustaPod]:
        return self.pod

    def get_deployment(self) -> Optional[RobustaDeployment]:
        return self.deployment

    def get_job(self) -> Optional[RobustaJob]:
        return self.job

    def get_daemonset(self) -> Optional[DaemonSet]:
        return self.daemonset

    def get_title(self) -> str:
        annotations = self.alert.annotations
        if annotations.get("summary"):
            return f'{annotations["summary"]}'
        else:
            return self.alert.labels.get("alertname", "")

    def get_prometheus_query(self) -> str:
        """
        Gets the prometheus query that defines this alert.
        """
        url = urlparse(self.alert.generatorURL)
        return re.match(r"g0.expr=(.*)&g0.tab=1", unquote_plus(url.query)).group(1)

    def get_description(self) -> str:
        annotations = self.alert.annotations
        clean_description = ""
        if annotations.get("description"):
            # remove "LABELS = map[...]" from the description as we already add a TableBlock with labels
            clean_description = re.sub(
                r"LABELS = map\[.*\]$", "", annotations["description"]
            )
        return clean_description

    def __get_alert_subject(self) -> FindingSubject:
        subject_type: FindingSubjectType = FindingSubjectType.TYPE_NONE
        name: Optional[str] = None
        namespace: Optional[str] = None
        node_name: Optional[str] = None
        if self.pod:
            subject_type = FindingSubjectType.TYPE_POD
            name = self.pod.metadata.name
            namespace = self.pod.metadata.namespace
            node_name = self.pod.spec.nodeName
        elif self.job:
            subject_type = FindingSubjectType.TYPE_JOB
            name = self.job.metadata.name
            namespace = self.job.metadata.namespace
        elif self.deployment:
            subject_type = FindingSubjectType.TYPE_DEPLOYMENT
            name = self.deployment.metadata.name
            namespace = self.deployment.metadata.namespace
        elif self.daemonset:
            subject_type = FindingSubjectType.TYPE_DAEMONSET
            name = self.daemonset.metadata.name
            namespace = self.daemonset.metadata.namespace
        elif self.node:
            subject_type = FindingSubjectType.TYPE_NODE
            name = self.node.metadata.name
            node_name = self.node.metadata.name

        return FindingSubject(name, subject_type, namespace, node_name)

    def create_default_finding(self) -> Finding:
        alert_subject = self.__get_alert_subject()
        status_message = '[RESOLVED] ' if self.alert.status.lower() == 'resolved' else ''
        title = f'{status_message}{self.get_title()}'
        return Finding(
            title=title,
            description=self.get_description(),
            source=FindingSource.PROMETHEUS,
            aggregation_key=self.alert_name,
            severity=SEVERITY_MAP.get(
                self.alert.labels.get("severity"), FindingSeverity.INFO
            ),
            subject=alert_subject,
        )
